(* Pdf-Viewer using Adobe Reader
   =============================

   2009-2023  Dr. J. Rathlev, (kontakt(a)rathlev-home.de)

   -----------------------------------------------------------------------------
   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Jun. 2009
   March 2017 (Delphi 10)
   last modified: Apr. 2023
   *)

unit AcroMain;

interface

uses Winapi.Windows, System.SysUtils, System.Classes, Vcl.Graphics, Vcl.Forms,
  Vcl.Controls, Vcl.Menus, Vcl.StdCtrls, Vcl.Dialogs, Vcl.Buttons, Winapi.Messages,
  Vcl.ExtCtrls, Vcl.ComCtrls, Vcl.StdActns,
  Vcl.ActnList, Vcl.ToolWin, Vcl.ImgList, HListBox;

const
  Prog = 'PDF-View';
  Vers = ' (Vers. 2.2)';
  CopRgt = ' 2009-2023 Dr. J. Rathlev, D-24222 Schwentinental';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  WM_SHOW = WM_USER+1987;

type
  TProcessString = procedure (const AStr : string) of object;

  TMainForm = class(TForm)
    OpenDialog: TOpenDialog;
    pnlTop: TPanel;
    bbFileList: TBitBtn;
    pmFiles: TPopupMenu;
    bbInfo: TBitBtn;
    pnlRight: TPanel;
    btnLast: TBitBtn;
    bbFirst: TBitBtn;
    pmWindows: TPopupMenu;
    itmVert: TMenuItem;
    itmHor: TMenuItem;
    bbWindows: TBitBtn;
    bbPreviousDoc: TBitBtn;
    bbNextDoc: TBitBtn;
    bbLoad: TBitBtn;
    bbClose: TBitBtn;
    bbUpdate: TBitBtn;
    bbExit: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure bbLastClick(Sender: TObject);
    procedure bbFirstClick(Sender: TObject);
    procedure bbLoadClick(Sender: TObject);
    procedure bbInfoClick(Sender: TObject);
    procedure bbExitClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure bbFileListClick(Sender: TObject);
    procedure bbWindowsClick(Sender: TObject);
    procedure bbPreviousDocClick(Sender: TObject);
    procedure bbNextDocClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure WMExitSizeMove(var Message: TWMNoParams); message WM_EXITSIZEMOVE;
    procedure itmHorClick(Sender: TObject);
    procedure itmVertClick(Sender: TObject);
    procedure bbCloseClick(Sender: TObject);
    procedure bbUpdateClick(Sender: TObject);
  private
    { Private declarations }
    ProgVersName,
    ProgVersDate,
    AppPath,UserPath,
    ProgPath,IniName,
    LastDir,LoadName : string;
    FileList,WinList : THistoryList;
    function CheckAcroPdf : boolean;
    procedure CreateMDIChild (const FName: string);
    procedure LoadPdfFile (const FileName : string);
    procedure LoadFromFileListClick (Sender : TObject; FileName : string);
    procedure ShowWindowClick (Sender : TObject; FileName : string);
    procedure MsgHandler (var Msg : TMsg; var Handled : boolean);
    procedure AppActivate (Sender : TObject);
//    function EnumChildProc(Handle: THandle; lParam: LongWord): Boolean; stdcall;
  public
    { Public declarations }
    procedure UpdateWinList(const FName : string);
    end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses AcroChild, System.Win.Registry, System.IniFiles, WinApiUtils, WinUtils, MsgDialogs,
  WinShell, Winapi.ShlObj, InpNumber, LangUtils, InitProg, GnuGetText, PathUtils;

{ ------------------------------------------------------------------- }
var
  mutex         : THandle;
  h,AppWin      : HWnd;
  Arg           : string;
  AppClass      : TWndClass;
  ProcessString : TProcessString;
  cds           : TCopyDataStruct;
  buf           : array [0..255] of char;

{ ------------------------------------------------------------------- }
const
  IniExt = 'ini';
  CfgSekt = 'Config';
  FileSekt = 'Files';

  IniLeft= 'Left';
  IniTop = 'Top';
  IniWidth = 'Width';
  IniHeight = 'Height';

  iniLastDir = 'LastDir';

// Prfe, ob eine geeignete Version von Adobe Reader oder Acrobat
// installiert ist
function TMainForm.CheckAcroPdf : boolean;
var
  Reg: TRegistry;
  AcroBasekey, AcroTypeLibID : string;
begin
  Result:=false;
  // If the "AcroBasekey" under HKEY_CLASSES_ROOT\CLSID\ exists,
  // an Adobe Reader or Acrobat version is installed.
  // The Value of the "TypeLib" Key determines the exact ActiveX version
  AcroBasekey := '{CA8A9780-280D-11CF-A24D-444553540000}';
  AcroTypeLibID := '{05BFD3F1-6319-4F30-B752-C7A22889BCC4}'; // v7ff

  // Check for installed acrobat reader
  Reg := TRegistry.Create(KEY_READ); // We only need read-only access
  try
    Reg.RootKey := HKEY_CLASSES_ROOT;
    if Reg.OpenKeyReadOnly('CLSID\'+AcroBasekey+'\TypeLib\') then begin
      // check TypeLib-ID
      Result:=Reg.ReadString('')=AcroTypeLibID;
      Reg.CloseKey;
      end
  finally
    Reg.Free;
    end;
  end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  IniFile  : TIniFile;
begin
  TranslateComponent(self);
  if not CheckAcroPdf then begin
    ErrorDialog(_('Adobe Reader or Acrobat not found!'));
    Application.Terminate;
    end;
  pnlTop.ParentBackground:=false;
  InitPaths(AppPath,UserPath,ProgPath);
  InitVersion(Prog,Vers,CopRgt,3,3,ProgVersName,ProgVersDate);
  Caption:='PDF View'+VersInfo.Comments;
  IniName:=Erweiter(AppPath,PrgName,IniExt);
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    LastDir:=ReadString(CfgSekt,iniLastDir,UserPath);
    Left:=ReadInteger (CfgSekt,IniLeft,Left);
    Top:=ReadInteger (CfgSekt,IniTop,Top);
    ClientWidth:=ReadInteger (CfgSekt,IniWidth,ClientWidth);
    ClientHeight:=ReadInteger (CfgSekt,IniHeight,ClientHeight);
    Free;
    end;
  Caption:=GetFileVersionName(Application.Exename,Prog,Vers);
// Message-Handler for application - see Mutex
  with Application do begin
    Title:=Prog;
    OnMessage:=MsgHandler;
    OnActivate:=AppActivate;
    end;
  (* expand file menu *)
  FileList:=THistoryList.Create;
  with FileList do begin
    LoadFromIni (IniName,FileSekt);
    Menu:=pmFiles.Items;
    OnAutoItemClick:=LoadFromFileListClick;
    end;
  WinList:=THistoryList.Create;
  with WinList do begin
    Menu:=pmWindows.Items;
    RadioMenu:=true;
    OnAutoItemClick:=ShowWindowClick;
    end;
  // open file from command line
  if ParamCount>0 then LoadName:=ParamStr(1) else LoadName:='';
  ProcessString:=LoadPdfFile;
  bbClose.Enabled:=false;
  end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
var
  i : integer;
begin
  for i:=MDIChildCount-1 downto 0 do MDIChildren[i].Close;
  with TIniFile.Create(IniName) do begin
    WriteString(CfgSekt,iniLastDir,LastDir);
    WriteInteger(CfgSekt,IniLeft,Left);
    WriteInteger(CfgSekt,IniTop,Top);
    WriteInteger(CfgSekt,IniWidth,ClientWidth);
    WriteInteger(CfgSekt,IniHeight,ClientHeight);
    Free;
    end;
  FileList.SaveToIni (IniName,FileSekt,true);
  end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FileList.Free;
  WinList.Free;
  end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  if length(LoadName)>0 then CreateMDIChild(LoadName);
  end;

procedure TMainForm.AppActivate (Sender : TObject);
begin
  if MDIChildCount>0 then with (ActiveMDIChild as TChildForm) do begin
    UpdateDoc;
    end;
  end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  if Visible and (MDIChildCount>0) then begin
    // Because of the strange behavior of the new Acrobat 7 ActiveX control,
    // it has to loose and gain focus again to repaint itself to the new size.
    // You have to give the focus to an control on the form, not the form itself.
    pnlTop.SetFocus; // Using the form here doesn't work
    with (ActiveMDIChild as TChildForm) do Pdf.SetFocus;
    end;
  end;

{ ------------------------------------------------------------------- }
procedure TMainForm.MsgHandler (var Msg : TMsg; var Handled : boolean);
begin
  if Msg.Message=WM_SHOW then begin
//    if not visible then Show;
    Application.BringToFront;
    Handled:=true;
//    if MDIChildCount>0 then (ActiveMDIChild as TChildForm).UpdateDoc;
    end;
  end;

{ ------------------------------------------------------------------- }
procedure TMainForm.CreateMDIChild (const FName: string);
begin
  //when user open a new documnet, create a new MDI child window
  if FileExists(FName) then begin
    WinList.AddString(FName);
    TChildForm.Create(self).Initialize(FName);
    FileList.AddString(FName);
    LastDir:=ExtractFilePath(FName);
    end
  else ErrorDialog (Format(_('File not found:'+sLineBreak+'%s'),[FName]));
  end;

// Load Pdf file
procedure TMainForm.LoadPdfFile (const FileName : string);
var
  i : integer;
begin
  if MDIChildCount>0 then for i:=0 to MDIChildCount-1 do begin
    if AnsiSameText((MDIChildren[i] as TChildForm).PdfName,FileName) then break;
    end
  else i:=1;
  if (i<MDIChildCount) then with MDIChildren[i] do begin
    BringToFront;
    UpdateWinList(FileName);
    end
  else CreateMDIChild(FileName);
  end;

procedure TMainForm.UpdateWinList(const FName : string);
begin
  WinList.SelectMenuItem(FName);
  bbClose.Enabled:=WinList.Count>1;
  end;

{ ------------------------------------------------------------------- }
procedure TMainForm.itmHorClick(Sender: TObject);
begin
  TileMode:=tbHorizontal;
  Tile;
  end;

procedure TMainForm.itmVertClick(Sender: TObject);
begin
  TileMode:=tbVertical;
  Tile;
  end;

procedure TMainForm.ShowWindowClick (Sender : TObject; FileName : string);
begin
  LoadPdfFile(FileName);
  end;

procedure TMainForm.bbFileListClick(Sender: TObject);
begin
  with CursorPos do pmFiles.Popup(x,y);
  end;

procedure TMainForm.bbWindowsClick(Sender: TObject);
begin
  with CursorPos do pmWindows.Popup(x,y);
  end;

{ ------------------------------------------------------------------- }
procedure TMainForm.bbLoadClick(Sender: TObject);
var
  i : integer;
begin
  with OpenDialog do begin
    if MDIChildCount>0 then InitialDir:=ExtractFilePath((ActiveMDIChild as TChildForm).PdfName)
    else InitialDir:=LastDir;
    if Execute then with Files do for i:=0 to Count-1 do LoadPdfFile(Strings[i]);
    end;
  end;

(* click on recent files *)
procedure TMainForm.LoadFromFileListClick (Sender     : TObject;
                                           FileName   : string);
begin
  if FileExists(FileName) then LoadPdfFile(FileName)
  else begin
    ErrorDialog(_('File not found!'));
    FileList.RemString(FileName);
    end;
  end;

procedure TMainForm.bbCloseClick(Sender: TObject);
var
  mdi : TChildForm;
begin
  if MDIChildCount>0 then begin
    mdi:=(ActiveMDIChild as TChildForm);
    Previous;
    WinList.RemString(mdi.PdfName);
    mdi.Close;
    UpdateWinList((ActiveMDIChild as TChildForm).PdfName);
    end;
  end;

procedure TMainForm.bbExitClick(Sender: TObject);
begin
  Close;
  end;

{ ------------------------------------------------------------------- }
procedure TMainForm.bbInfoClick(Sender: TObject);
begin
  InfoDialog(CursorPos,ProgVersName+' - '+ProgVersDate+sLineBreak+CopRgt
           +sLineBreak+'E-Mail: '+EmailAdr);
  end;

procedure TMainForm.bbPreviousDocClick(Sender: TObject);
begin
  Previous;
  end;

procedure TMainForm.bbNextDocClick(Sender: TObject);
begin
  Next;
  end;

procedure TMainForm.bbUpdateClick(Sender: TObject);
begin
  if MDIChildCount>0 then (ActiveMDIChild as TChildForm).Reload;
  end;

{ ------------------------------------------------------------------- }
procedure TMainForm.bbFirstClick(Sender: TObject);
begin
  if MDIChildCount>0 then (ActiveMDIChild as TChildForm).GotoFirstPage;
  end;

procedure TMainForm.bbLastClick(Sender: TObject);
begin
  if MDIChildCount>0 then (ActiveMDIChild as TChildForm).GotoLastPage;
  end;

procedure TMainForm.WMExitSizeMove(var Message: TWMNoParams);
begin
  if MDIChildCount>0 then (ActiveMDIChild as TChildForm).Reload;
  end;

{ ------------------------------------------------------------------- }
function IMWndProc (h : hWnd; Msg,wParam,lParam : longint) : longint; stdcall;
var
  s : string;
begin
  if Msg=WM_COPYDATA then begin
    s:=PChar(PCopyDataStruct(lParam).lpData);
    if assigned(ProcessString) then ProcessString(s);
    Result:=ord(true);
    end
  else Result:=DefWindowProc(h,Msg,wParam,lParam);
  end;

{ ------------------------------------------------------------------- }
initialization
  Application.Title:=Prog;
  ProcessString:=nil;
  AppWin:=0;
  with AppClass do begin
    lpfnWndProc:=@IMWndProc;
    hInstance:=hInstance;
    lpszClassName:='TInstanceManager';
    end;
  if Winapi.Windows.RegisterClass (AppClass)=0 then RaiseLastOSError;
  mutex := CreateMutex(nil,true,Prog);
  if (Mutex = 0) or (getLastError=ERROR_ALREADY_EXISTS) then begin
    if ParamCount>0 then Arg:=ParamStr(1) else Arg:='';
    if (length(Arg)>0) and (WaitForSingleObject (mutex,5000)=WAIT_OBJECT_0) then begin
      AppWin:=FindWindow('TInstanceManager',Prog);
      if AppWin=0 then RaiseLastOSError
      else begin
        with cds do begin     // Nachrichtenblock
          dwData:=1; cbData:=SizeOf(buf);
          lpData:=StrPCopy(buf,Arg);
          end;
        Winapi.Windows.SendMessage (AppWin,WM_COPYDATA,0,integer(@cds));
        end;
      AppWin:=0;
      end;
    h:=0;
    repeat
      h:=FindWindowEx(0,h,'TApplication',PChar(Application.Title))
      until h<>application.handle;
    if h<>0 then begin
      if IsWindowVisible(h) then begin
        (* Fenster in den Vordergrund *)
        Winapi.Windows.ShowWindow(h,SW_ShowNormal);
        Winapi.Windows.SetForegroundWindow(h);
        end
      else (* Fenster anzeigen *)
        Winapi.Windows.PostMessage (h,WM_SHOW,0,0);
      end;
    ReleaseMutex(mutex);
    halt;
    end
  else begin    // starte 1. Instanz
    AppWin:=CreateWindow(AppClass.lpszClassName,Prog,0,0,0,0,0,0,0,hInstance,nil);
    ReleaseMutex(mutex);
    end;

finalization
  if AppWin>0 then DestroyWindow(AppWin);
end.
